from dataclasses import dataclass
import os
from typing import Dict, Tuple

import numpy as np


"""
A template of an INTERFILE header.
"""
TEMPLATE_HEADER_IMAGE = """
INTERFILE  :=
imaging modality := nucmed
version of keys := STIR4.0
name of data file := {DATA_FILE}

GENERAL DATA :=

GENERAL IMAGE DATA :=
type of data := Tomographic
isotope name := ^99m^Technetium

patient orientation := {PATIENT_ORIENTATION}
patient rotation := {PATIENT_ROTATION}

imagedata byte order := LITTLEENDIAN

SPECT STUDY (General) :=
process status := Reconstructed

number format := float
number of bytes per pixel := 4
number of dimensions := 3

matrix axis label [1] := x
matrix size [1] := {ROWS}
scaling factor (mm/pixel) [1] := {SPACING_X}
matrix axis label [2] := y
matrix size [2] := {COLUMNS}
scaling factor (mm/pixel) [2] := {SPACING_Y}
matrix axis label [3] := z
matrix size [3] := {SLICES}
scaling factor (mm/pixel) [3] := {SPACING_Z}

first pixel offset (mm) [1] := {OFFSET_X}
first pixel offset (mm) [2] := {OFFSET_Y}
first pixel offset (mm) [3] := 0

number of time frames := 1
END OF INTERFILE :=
"""


"""
Default extension for the Interfile header.
"""
EXTENSION_INTERFILE_HEADER = ".hv"
"""
Default extension for the Interfile image.
"""
EXTENSION_INTERFILE_IMAGE = ".v"


@dataclass
class Interfile:
    """
    Interfile data type wrapping the header and the image
    into an object.
    """

    header: Dict[str, str]
    image: np.ndarray

    def __iter__(self):
        return iter((self.header, self.image))


@dataclass
class _InterfileKeys:
    """
    Data class defining keys for an Interfile header.
    """

    placeholder: str = "{_}"

    _dim: str = f"matrix size [{placeholder}]"
    _spacing: str = f"scaling factor (mm/pixel) [{placeholder}]"

    data_file: str = "name of data file"
    n_projections: str = "number of projections"
    bytes_per_pixel: str = "number of bytes per pixel"
    number_format: str = "number format"

    patient_orientation: str = "patient orientation"
    patient_rotation: str = "patient rotation"

    def dim(self, index: int) -> str:
        return self._dim.replace(self.placeholder, str(index))

    def spacing(self, index: int) -> str:
        return self._spacing.replace(self.placeholder, str(index))


InterfileKeys = _InterfileKeys()


def type_by_format(number_format: str, bytes_per_pixel: int) -> type:
    """
    Get the corresponding numpy array type for a number format and bytes per
    pixel as defined in an INTERFILE header.

    :param number_format: the number format as a string, e.g. "float"
    :param bytes_per_pixel: the amount of bytes stored for each pixel
    :return: a corresponding numpy data type
    """
    if number_format == "float" and bytes_per_pixel == 4:
        return np.float32

    raise ValueError(
        "Unknown mapping from format {number_format} with {bytes_per_pixel} bytes to numpy type"
    )


def parse_interfile_header_str(header: str) -> Dict[str, str]:
    """
    Parse an INTERFILE header into a dict.
    This is done by splitting non-empty lines in the INTERFILE header by ":=".

    :param header: the text of an INTERFILE header
    :return: a dictionary of value in the header
    """
    header = header.replace("!", "")
    lines = header.strip().split("\n")
    items = map(lambda line: line.split(":="), lines)
    items = filter(lambda item: len(item) == 2, items)
    items = map(lambda item: [item[0].strip(), item[1].strip()], items)
    return dict(items)


def parse_interfile_header(filename: str) -> Dict[str, str]:
    """
    Parse an INTERFILE header into a dict.
    This is done by splitting non-empty lines in the INTERFILE header by ":=".

    :param filename: the filename of the INTERFILE header
    :return: a dictionary of value in the header
    """
    with open(filename, mode="r") as f:
        return parse_interfile_header_str(f.read())


def load_interfile(filename: str) -> Interfile:
    """
    Load an INTERFILE header and its image as a numpy array.

    :param filename: the filename of the INTERFILE header file
    :return: the interfile image
    """
    header = parse_interfile_header(filename)

    dim_x = int(header[InterfileKeys.dim(1)])
    dim_y = int(header[InterfileKeys.dim(2)])
    dim_z = (
        int(header[InterfileKeys.dim(3)])
        if InterfileKeys.dim(3) in header
        else int(header[InterfileKeys.n_projections])
    )

    bytes_per_pixel = int(header[InterfileKeys.bytes_per_pixel])
    num_format = header[InterfileKeys.number_format]
    dtype = type_by_format(num_format, bytes_per_pixel)

    data_file = os.path.join(os.path.dirname(filename), header[InterfileKeys.data_file])
    with open(data_file, mode="rb") as f:
        image = np.frombuffer(f.read(), dtype)
    image = image.reshape((dim_z, dim_y, dim_x))
    return Interfile(header, image.copy())


def load_interfile_img(filename: str) -> np.ndarray:
    """
    Load an INTERFILE image as a numpy array.

    :param filename: the filename of the INTERFILE header file
    :return: the image as a numpy array
    """
    interfile = load_interfile(filename)
    return interfile.image


def write_interfile(filename: str, interfile: Interfile):
    """
    Write an interfile image.
    Note that custom extensions to the filename are removed and the defaults
    from this module are used (.hv for the header and .v for the image).
    This method also sanitizes the header by making sure that
    dimensions match the image.

    :param filename: the filename where the interfile image is written to
    :param interfile: the interfile saved
    """
    filename = os.path.splitext(filename)[0]
    filename_data = f"{filename}{EXTENSION_INTERFILE_IMAGE}"
    filename_header = f"{filename}{EXTENSION_INTERFILE_HEADER}"

    header, image = interfile

    header[InterfileKeys.data_file] = os.path.basename(filename_data)
    header[InterfileKeys.dim(3)] = str(image.shape[0])
    header[InterfileKeys.dim(2)] = str(image.shape[1])
    header[InterfileKeys.dim(1)] = str(image.shape[2])

    image = image.astype(np.float32)

    header_str = map(lambda item: f"{item[0]} := {item[1]}", header.items())
    header_str = "\n".join(header_str)
    with open(filename_header, mode="w") as f:
        f.write(header_str)
    with open(filename_data, mode="wb") as f:
        f.write(image.tobytes())


if __name__ == "__main__":
    import argparse
    import math

    parser = argparse.ArgumentParser(
        description="Modify an interfile image",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--interfile", type=str, required=True, help="the interfile input"
    )
    parser.add_argument(
        "--out",
        type=str,
        help="the interfile output - if not set the input file will be overwritte",
    )
    parser.add_argument(
        "--cmd",
        choices=["fill", "crop", "pad", "flip"],
        help="the modification to perform",
    )
    parser.add_argument(
        "--param",
        type=int,
        required=True,
        help="the parameters for the modification [fill: value, crop & pad: target size of z dimension]",
    )
    args = parser.parse_args()
    args.out = args.interfile if args.out is None else args.out

    interfile = load_interfile(args.interfile)

    if args.cmd == "fill":
        value = args.param
        interfile.image.fill(value)
    if args.cmd == "pad":
        target_size = args.param
        real_size = interfile.image.shape[0]
        assert (
            target_size > real_size
        ), f"Cannot pad from a larger size {real_size} to a smaller size {target_size}"

        diff = target_size - real_size
        pad_lower = math.ceil(diff / 2)
        pad_upper = math.floor(diff / 2)
        interfile.image = np.pad(
            interfile.image, ((pad_lower, pad_upper), (0, 0), (0, 0))
        )
    if args.cmd == "crop":
        target_size = args.param
        real_size = interfile.image.shape[0]
        assert (
            target_size < real_size
        ), f"Cannot crop from a smaller size {real_size} to a larger size {target_size}"

        diff = target_size / 2
        center = real_size // 2
        crop_lower = center - math.floor(diff)
        crop_upper = center + math.ceil(diff)
        interfile.image = interfile.image[crop_lower:crop_upper]
    if args.cmd == "flip":
        interfile.image = interfile.image[:, :, ::-1]

    write_interfile(args.out, interfile)
